"""
The core algorithm of RNA-Sieve, which is defined as `find_mixtures(...)`.
"""

import numpy as np
import cvxpy as cp
import scipy.optimize
import math
from multiprocessing import Pool
from multiprocessing import sharedctypes
from rnasieve.helper import CLIP_VALUE, ALPHA_EPS, ALPHA_MAX_ITER, compute_mixture_sigma, compute_weighted_norm, compute_full_likelihood, compute_row_likelihood

# Parallelization Global shared arrays

shared_dict = {}

# Updates (1)


def _update_n(phi, alpha, psi, sigma):
    mixture_variance = np.clip(compute_mixture_sigma(
        alpha, sigma, phi), CLIP_VALUE, None)
    G = psi.shape[0]
    psi_hat_term = compute_weighted_norm(phi @ alpha.T, 1 / mixture_variance)
    observed_psi_term = compute_weighted_norm(psi, 1 / mixture_variance)
    return np.max(np.roots([psi_hat_term, G, -observed_psi_term]))


def _minimize_phi_row(
        phi_prev,
        psi_scalars,
        observed_phi_row,
        sigma_row,
        alphas,
        ns,
        m,
        row_idx=None,
        parallelized=False,
        shared_array_id=None):
    phi_next = cp.Variable(phi_prev.shape)
    phi_final = np.copy(phi_prev)
    try:
        phi_prev = np.clip(phi_prev, CLIP_VALUE, None)
        sample_coef = m / np.clip(sigma_row, CLIP_VALUE, None)
        bulk_coef = ns / \
            np.clip(compute_mixture_sigma(
                alphas, sigma_row, phi_prev), CLIP_VALUE, None)
        prob = cp.Problem(
            cp.Minimize((observed_phi_row - phi_next)**2 @ sample_coef.T +
                        (phi_next @ alphas.T - psi_scalars / ns)**2 @ bulk_coef.T),
            [phi_next >= 0])
        prob.solve()
        phi_final = np.clip(phi_next.value, 0, None)
    except (cp.SolverError, TypeError) as e:
        pass
    if phi_next.value is None:
        phi_final = np.clip(phi_prev, 0, None)

    if parallelized and shared_array_id:
        phi_tmp = np.ctypeslib.as_array(shared_dict[shared_array_id])
        phi_tmp[row_idx, :] = phi_final

    return phi_final


def _minimize_phi(
        phi_prev,
        psis,
        observed_phi,
        sigma,
        alphas,
        ns,
        m,
        parallelized=False,
        num_process=1):
    if not parallelized:
        phi_next = np.empty((0, phi_prev.shape[1]), dtype=np.float64)
        for i in range(phi_prev.shape[0]):
            phi_next = np.vstack((phi_next, _minimize_phi_row(
                phi_prev[i, :].reshape(1, -1),
                psis[i].reshape(1, -1),
                observed_phi[i, :].reshape(1, -1),
                sigma[i, :].reshape(1, -1),
                alphas, ns, m)))
        return phi_next
    else:
        phi_next = np.ctypeslib.as_ctypes(np.zeros(phi_prev.shape))
        phi_next_shared_array = sharedctypes.RawArray(
            phi_next._type_, phi_next)
        shared_array_id = "MINIMIZE_PHI"
        shared_dict[shared_array_id] = phi_next_shared_array
        p = Pool(processes=num_process)
        p.starmap(_minimize_phi_row, [(
            phi_prev[i, :].reshape(1, -1),
            psis[i].reshape(1, -1),
            observed_phi[i, :].reshape(1, -1),
            sigma[i, :].reshape(1, -1),
            alphas, ns, m, i, True, shared_array_id) for i in range(phi_prev.shape[0])])
        p.close()
        return np.ctypeslib.as_array(phi_next_shared_array)


def _minimize_alpha(alpha_prev, sigma, phi, psi, n):
    alpha_next = cp.Variable(alpha_prev.shape)
    alpha = np.copy(alpha_prev)
    try:
        coef = n / np.clip(compute_mixture_sigma(alpha_prev,
                                                 sigma, phi), CLIP_VALUE, None)
        prob = cp.Problem(cp.Minimize(
            coef.T @ (phi @ alpha_next.T - psi / n)**2,
            [alpha_next >= 0, cp.sum(alpha_next) == 1]))
        prob.solve()
        alpha = np.clip(alpha_next.value, 0, None)
    except (cp.SolverError, TypeError) as e:
        pass
    if alpha_next.value is None:
        return alpha_prev
    return alpha / np.sum(alpha)


def _minimize_alpha_LS_helper(alpha_prev, sigma, phi, psi, n, eps, max_iter):
    alpha_last = alpha_prev
    alpha_cur = _minimize_alpha(alpha_prev, sigma, phi, psi, n)
    it = 0
    while np.linalg.norm(alpha_cur - alpha_last,
                         ord=2) > eps and it < max_iter:
        alpha_cur, alpha_last = _minimize_alpha(
            alpha_cur, sigma, phi, psi, n), alpha_cur
        it += 1
    return alpha_cur


def _alternate_minimization(
        phi,
        sigma,
        m,
        psis,
        alpha_inits,
        n_inits,
        eps,
        delta,
        max_iter,
        parallelized=False,
        num_process=1):
    L_past = compute_full_likelihood(
        phi, phi, psis, alpha_inits, sigma, n_inits, m)
    phi_next = _minimize_phi(
        phi,
        psis,
        phi,
        sigma,
        alpha_inits,
        n_inits,
        m,
        parallelized,
        num_process)
    alpha_nexts = np.zeros(alpha_inits.shape)
    for i in range(alpha_nexts.shape[0]):
        alpha_nexts[i] = _minimize_alpha_LS_helper(alpha_inits[i].reshape(
            1, -1), sigma, phi_next, psis[:, i].reshape(-1, 1), n_inits[i],
            ALPHA_EPS, ALPHA_MAX_ITER)
    n_nexts = np.zeros(n_inits.shape)
    for i in range(n_nexts.shape[0]):
        n_nexts[i] = _update_n(phi_next, alpha_nexts[i].reshape(
            1, -1), psis[:, i].reshape(-1, 1), sigma)
    L_next = compute_full_likelihood(
        phi_next, phi, psis, alpha_nexts, sigma, n_nexts, m)

    if L_next + delta > L_past:
        alpha_nexts = alpha_inits

    phi_past = phi
    alpha_pasts = alpha_inits.copy()
    n_pasts = n_inits.copy()

    it = 0
    while (
            np.max(
                np.linalg.norm(
                    alpha_nexts - alpha_pasts,
                    ord=np.inf,
                    axis=1)) > eps and it < max_iter and L_next +
            delta < L_past):
        phi_past, alpha_pasts, n_pasts, L_past = phi_next, alpha_nexts, n_nexts, L_next

        phi_next = _minimize_phi(phi_past, psis, phi, sigma, alpha_pasts,
                                 n_nexts, m, parallelized, num_process)
        alpha_nexts = np.zeros(alpha_pasts.shape)
        for i in range(alpha_nexts.shape[0]):
            alpha_nexts[i] = _minimize_alpha_LS_helper(alpha_pasts[i].reshape(1, -1),
                                                       sigma, phi_next,
                                                       psis[:, i].reshape(-1, 1),
                                                       n_pasts[i],
                                                       ALPHA_EPS, ALPHA_MAX_ITER)
        n_nexts = np.zeros(n_pasts.shape)
        for i in range(n_nexts.shape[0]):
            n_nexts[i] = _update_n(phi_next, alpha_nexts[i].reshape(
                1, -1), psis[:, i].reshape(-1, 1), sigma)
        L_next = compute_full_likelihood(
            phi_next, phi, psis, alpha_nexts, sigma, n_nexts, m)

        it += 1

    return alpha_nexts, n_nexts, phi_next

# Updates (2)


def _minimize_phi_row_grad(
        phi_past,
        observed_phi_row,
        alphas,
        ns,
        observed_psi_scalars,
        sigma_row,
        m,
        row_idx=None,
        parallelized=False,
        shared_array_id=None):
    bnds = tuple((0, None) for _ in range(phi_past.shape[1]))
    res = scipy.optimize.minimize(
        lambda phi_row: compute_row_likelihood(
            phi_row.reshape(1, -1),
            observed_phi_row,
            observed_psi_scalars,
            alphas,
            sigma_row,
            ns,
            m),
        phi_past,
        method='SLSQP',
        bounds=bnds)
    phi_row_clipped = np.clip(res.x.reshape(1, -1), 0., None)
    phi_row_clipped[phi_row_clipped <= CLIP_VALUE] = 0.

    if parallelized and shared_array_id:
        phi_tmp = np.ctypeslib.as_array(shared_dict[shared_array_id])
        phi_tmp[row_idx, :] = phi_row_clipped

    return phi_row_clipped


def _minimize_phi_grad(
        phi_past,
        observed_phi,
        alphas,
        ns,
        observed_psis,
        sigma,
        m,
        parallelized=False,
        num_process=1):
    if not parallelized:
        phi_next = np.empty((0, phi_past.shape[1]), dtype=np.float64)
        for i in range(phi_past.shape[0]):
            phi_next = np.vstack((phi_next, _minimize_phi_row_grad(
                phi_past[i, :].reshape(1, -1),
                observed_phi[i, :].reshape(1, -1),
                alphas,
                ns,
                observed_psis[i].reshape(1, -1),
                sigma[i, :].reshape(1, -1),
                m)))
        return phi_next
    else:
        phi_next = np.ctypeslib.as_ctypes(np.zeros(phi_past.shape))
        phi_next_shared_array = sharedctypes.RawArray(
            phi_next._type_, phi_next)
        shared_array_id = "PHI_GRAD"
        shared_dict[shared_array_id] = phi_next_shared_array
        p = Pool(processes=num_process)
        p.starmap(_minimize_phi_row_grad, [
            (phi_past[i, :].reshape(1, -1),
             observed_phi[i, :].reshape(1, -1),
             alphas,
             ns,
             observed_psis[i].reshape(1, -1),
             sigma[i, :].reshape(1, -1),
             m,
             i,
             True,
             shared_array_id) for i in range(phi_past.shape[0])])
        p.close()
        return np.ctypeslib.as_array(phi_next_shared_array)


def _alternate_gradient_descent(
        phi,
        phi_init,
        sigma,
        m,
        psis,
        alpha_inits,
        n_inits,
        eps,
        delta,
        max_iter,
        parallelized=False,
        num_process=1):
    phi_past = phi_init
    alpha_pasts = alpha_inits
    n_pasts = n_inits
    L_past = compute_full_likelihood(
        phi_past, phi, psis, alpha_pasts, sigma, n_pasts, m)
    phi_next = _minimize_phi_grad(
        phi_past,
        phi,
        alpha_pasts,
        n_pasts,
        psis,
        sigma,
        m,
        parallelized,
        num_process)
    alpha_nexts = np.zeros(alpha_pasts.shape)
    for i in range(alpha_pasts.shape[0]):
        alpha_nexts[i] = _minimize_alpha_LS_helper(
            alpha_pasts[i].reshape(1, -1),
            sigma,
            phi_next,
            psis[:, i].reshape(-1, 1),
            n_pasts[i],
            ALPHA_EPS,
            ALPHA_MAX_ITER)
    n_nexts = np.zeros(n_pasts.shape)
    for i in range(n_nexts.shape[0]):
        n_nexts[i] = _update_n(phi_next, alpha_nexts[i].reshape(
            1, -1), psis[:, i].reshape(-1, 1), sigma)
    L_next = compute_full_likelihood(
        phi_next, phi, psis, alpha_nexts, sigma, n_nexts, m)

    if L_next + delta > L_past:
        alpha_nexts = alpha_pasts

    it = 0
    while np.max(np.linalg.norm(alpha_nexts - alpha_pasts,
                                ord=np.inf, axis=1)) > eps and it < max_iter:
        phi_past, alpha_pasts, n_pasts, L_past = phi_next, alpha_nexts, n_nexts, L_next

        phi_next = _minimize_phi_grad(
            phi_past,
            phi,
            alpha_pasts,
            n_pasts,
            psis,
            sigma,
            m,
            parallelized,
            num_process)
        alpha_nexts = np.zeros(alpha_pasts.shape)
        for i in range(alpha_pasts.shape[0]):
            alpha_nexts[i] = _minimize_alpha_LS_helper(
                alpha_pasts[i].reshape(1, -1),
                sigma,
                phi_next,
                psis[:, i].reshape(-1, 1),
                n_pasts[i],
                ALPHA_EPS,
                ALPHA_MAX_ITER)
        n_nexts = np.zeros(n_pasts.shape)
        for i in range(n_nexts.shape[0]):
            n_nexts[i] = _update_n(phi_next, alpha_nexts[i].reshape(
                1, -1), psis[:, i].reshape(-1, 1), sigma)
        L_next = compute_full_likelihood(
            phi_next, phi, psis, alpha_nexts, sigma, n_nexts, m)

        it += 1

    return alpha_nexts, n_nexts, phi_next


def _compute_alpha_LS(alpha_hats, phi_hat, phi, sigma, psis):
    alpha_LS = np.zeros(alpha_hats.shape)
    for i in range(alpha_hats.shape[0]):
        mixture_sigma_diag = np.diag(
            1 / np.sqrt(compute_mixture_sigma(alpha_hats[i].reshape(1, -1),
                                              sigma, phi_hat)).ravel())
        # Note: rcond parameter is set to silence a deprecation warning
        alpha_LS_i = np.linalg.lstsq(
            mixture_sigma_diag @ phi, mixture_sigma_diag @ psis[:, i].reshape(-1, 1),
            rcond=-1)[0].T
        alpha_LS_i = np.clip(alpha_LS_i, 0, None)
        alpha_LS[i] = alpha_LS_i / np.sum(alpha_LS_i)
    return alpha_LS


def find_mixtures(phi, sigma, m, psis, eps=1e-1, delta=1e-1, max_iter=10,
                  uniform_init=False, parallelized=True, num_process=10):
    """Infers alpha (mixture proportion), n (bulk cell count), and phi (mean matrix)
    given observed phi (mean matrix), sigma (variance matrix), m from a reference set,
    and psis, the bulks being deconvolved."""
    # Increment sigma by one to reduce numerical instability and zero inflation
    sigma += 1
    if uniform_init:
        alpha_inits = np.ones(phi.shape[1], 1) / phi.shape[1]
        n_inits = np.sum(phi @ alpha_inits.T, axis=0) / np.sum(psis, axis=0)
    else:
        # Note: rcond parameter is set to silence a deprecation warning
        naive_LS = np.clip(np.linalg.lstsq(phi, psis, rcond=-1)[0], 0, None)
        alpha_inits = naive_LS.T / np.sum(naive_LS, axis=0).reshape(-1, 1)
        n_inits = np.linalg.norm(naive_LS, ord=1, axis=0)
    L_init = compute_full_likelihood(
        phi, phi, psis, alpha_inits, sigma, n_inits, m)

    alpha_opts = alpha_inits.copy()
    n_opts = n_inits.copy()
    phi_opt = phi.copy()
    L_opt = L_init

    alpha_proj_opts = alpha_inits.copy()
    n_proj_opts = n_inits.copy()
    phi_proj_opt = phi.copy()
    L_proj_opt = L_init

    alpha_nexts, n_nexts, phi_hat = _alternate_minimization(
        phi, sigma, m, psis, alpha_inits, n_inits,
        eps, delta, max_iter, parallelized, num_process)
    alpha_nexts, n_nexts, phi_hat = _alternate_gradient_descent(
        phi, phi_hat, sigma, m, psis, alpha_nexts, n_nexts,
        eps, delta, max_iter, parallelized, num_process)
    L_pre = compute_full_likelihood(
        phi_hat, phi, psis, alpha_nexts, sigma, n_nexts, m)

    alpha_LS = _compute_alpha_LS(alpha_nexts, phi_hat, phi, sigma, psis)
    L_LS = compute_full_likelihood(
        phi_hat, phi, psis, alpha_LS, sigma, n_nexts, m)

    it = 0
    while L_LS < max(L_proj_opt, L_pre) and it < max_iter:
        if L_LS < L_proj_opt:
            alpha_proj_opts = alpha_LS.copy()
            n_proj_opts = n_nexts.copy()
            phi_proj_opt = phi_hat.copy()
            L_proj_opt = L_LS

        if L_pre < L_opt:
            alpha_opts = alpha_nexts.copy()
            n_opts = n_nexts.copy()
            phi_opt = phi_hat.copy()
            L_opt = L_pre

        alpha_nexts, n_nexts, phi_hat = _alternate_gradient_descent(
            phi, phi_hat, sigma, m, psis, alpha_LS, n_nexts,
            eps, delta, max_iter, parallelized, num_process)
        L_pre = compute_full_likelihood(
            phi_hat, phi, psis, alpha_nexts, sigma, n_nexts, m)

        alpha_LS = _compute_alpha_LS(alpha_nexts, phi_hat, phi, sigma, psis)
        L_LS = compute_full_likelihood(
            phi_hat, phi, psis, alpha_LS, sigma, n_nexts, m)

        it += 1

    if L_proj_opt < L_opt:
        alpha_opts = alpha_proj_opts.copy()
        n_opts = n_proj_opts.copy()
        phi_opt = phi_proj_opt.copy()
        L_opt = L_proj_opt

    return (alpha_proj_opts, n_proj_opts,
            phi_proj_opt), (alpha_opts, n_opts, phi_opt)



"""
Post-processing statistics calculations for RNA-Sieve.

Currently only contains functions for confidence interval computation.
"""
import numpy as np
import scipy.linalg
from scipy.stats import norm
from rnasieve.helper import compute_mixture_sigma, CLIP_VALUE

# Confidence Interval Computation


def _partial_mixture_variance_alpha(phi, sigma, alpha):
    return sigma + phi**2 - 2 * phi * (phi @ alpha.T)


def _partial_mixture_variance_phi(phi, alpha):
    return 2 * np.tile(alpha, (phi.shape[0], 1)) * (phi - phi @ alpha.T)


def _partial_alpha_alpha(phi, sigma, alpha, n):
    mixture_variance = compute_mixture_sigma(alpha, sigma, phi)
    partial_mv_alpha = _partial_mixture_variance_alpha(phi, sigma, alpha)
    return n * phi.T @ np.diag(1 / mixture_variance.reshape(-1)) @ phi + \
        partial_mv_alpha.T @ np.diag(
            1 / mixture_variance.reshape(-1)**2) @ partial_mv_alpha / 2


def _partial_phi_phi(phi, sigma, alpha, n, m):
    G, K = phi.shape
    mixture_variance = compute_mixture_sigma(alpha, sigma, phi)
    alpha_term = np.repeat(n * np.kron(alpha, alpha).reshape(K, K)[:, :, np.newaxis],
                           G, axis=2) / mixture_variance.reshape(1, 1, -1)
    partial_mv_phi = _partial_mixture_variance_phi(phi, alpha)
    phi_term = np.dstack([np.kron(partial_mv_phi[i], partial_mv_phi[i]).reshape(
        K, K) / (2 * mixture_variance[i]**2) for i in range(G)])
    return scipy.linalg.block_diag(*[np.squeeze(subarr) for subarr in np.dsplit(
        alpha_term + phi_term, G)]) + np.diag((m / sigma).reshape(-1))


def _partial_alpha_phi(phi, sigma, alpha, n):
    G, K = phi.shape
    mixture_variance = compute_mixture_sigma(alpha, sigma, phi)
    partial_mv_alpha = _partial_mixture_variance_alpha(phi, sigma, alpha)
    partial_mv_phi = _partial_mixture_variance_phi(phi, alpha)

    @np.vectorize
    def partial_alpha_phi_helper(i, j):
        g, k = j // K, j % K
        return n * alpha[0, k] * phi[g, i] / mixture_variance[g, 0] + \
            partial_mv_alpha[g, i] * partial_mv_phi[g, k] / \
            (2 * mixture_variance[g, 0]**2)

    return partial_alpha_phi_helper(*np.mgrid[0:K, 0:G * K])


def _partial_n_n(phi, sigma, alpha, n):
    mixture_variance = compute_mixture_sigma(alpha, sigma, phi)
    G = phi.shape[0]
    return (1 - G / 2) / n**2 + \
        np.sum((phi @ alpha.T)**2 / mixture_variance) / n


def _partial_alpha_n(phi, sigma, alpha, n):
    mixture_variance = compute_mixture_sigma(alpha, sigma, phi)
    partial_mv_alpha = _partial_mixture_variance_alpha(phi, sigma, alpha)
    return (phi / mixture_variance).T @ (phi @ alpha.T) + \
        np.sum(partial_mv_alpha / mixture_variance,
               axis=0).reshape(-1, 1) / (2 * n)


def _partial_n_phi(phi, sigma, alpha, n):
    mixture_variance = compute_mixture_sigma(alpha, sigma, phi)
    partial_mv_phi = _partial_mixture_variance_phi(phi, alpha)
    return (np.kron((phi @ alpha.T) / mixture_variance, alpha) +
            partial_mv_phi / (2 * n * mixture_variance)).reshape(1, -1)


def inverse_observed_fisher(phi, sigma, m, alpha, n):
    """Finds the inverse observed fisher information matrix.
    Returns an abbreviated square matrix of size K + 1
    where K is the number of cell types, to include the values
    for the inferred alpha and n values."""
    p_alpha_alpha = _partial_alpha_alpha(phi, sigma, alpha, n)
    p_n_n = _partial_n_n(phi, sigma, alpha, n)
    p_phi_phi = _partial_phi_phi(phi, sigma, alpha, n, m)
    p_alpha_phi = _partial_alpha_phi(phi, sigma, alpha, n)
    p_alpha_n = _partial_alpha_n(phi, sigma, alpha, n)
    p_n_phi = _partial_n_phi(phi, sigma, alpha, n)

    fisher_info = np.block([
        [p_alpha_alpha, p_alpha_n, p_alpha_phi],
        [p_alpha_n.T, p_n_n, p_n_phi],
        [p_alpha_phi.T, p_n_phi.T, p_phi_phi],
    ])

    inv_fisher = np.linalg.inv(fisher_info)
    return ((inv_fisher + inv_fisher.T) /
            2)[:alpha.shape[1] + 1, :alpha.shape[1] + 1]


def cross_protocol_inverse_observed_fisher(phi, sigma, m, alpha, n, psi):
    """Finds an adjusted inverse observed fisher information matrix
    for cross protocol experiments.

    Takes in phi, sigma, psi matrices filtered
    by filter_droplet_to_facs/filter_facs_to_droplet."""
    G = phi.shape[0]
    mixture_variance = compute_mixture_sigma(alpha, sigma, phi)
    z_scores = (phi @ alpha.T - psi / n) / np.sqrt(mixture_variance / n)
    z_tail = (z_scores**2).sum() - G
    filter_length = max(np.ceil(500 - 400 * max(z_tail, 0)), 100)
    filter_idxs = np.arange(G) if G < 500 else np.random.choice(
        G, min(filter_length, G), replace=False)

    return inverse_observed_fisher(phi[filter_idxs], np.clip(
        sigma[filter_idxs], 1, None), m, alpha, n)


def compute_marginal_confidence_intervals(
        phi, sigma, m, alpha, n, psi, sig=.05):
    """Computes an array of tuples, each representing a confidence interval of a
    cell type proportion estimate in alpha at a significance level `sig`."""
    inverse_obs_FI = cross_protocol_inverse_observed_fisher(
        phi, sigma, m, alpha, n, psi)
    c = norm.isf(sig / 2)
    return [(alpha[0, i] - c * np.sqrt(inverse_obs_FI[i, i]), alpha[0, i] +
             c * np.sqrt(inverse_obs_FI[i, i])) for i in range(alpha.shape[1])]



"""
Defines a class that can be used to run RNA-Sieve in an object-oriented manner.
"""
import numpy as np
import pandas as pd
import altair as alt
from rnasieve.algo import find_mixtures
from rnasieve.stats import compute_marginal_confidence_intervals as compute_marginal_cis
import numbers


class RNASieveModel:
    """
    RNASieveModel is a class that wraps the core algorithm of RNA-Sieve, and provides
    additional instance functions for plotting and computing confidence intervals.

    Initialize with reference data (phi, sigma, m) and run predict on bulk vector(s).
    """

    def __init__(self, observed_phi, observed_sigma, observed_m):
        assert observed_phi.shape == observed_sigma.shape, 'Reference mean matrix and variance matrix must have the same dimensions'
        assert observed_m.shape[1] == observed_phi.shape[1], 'Number of cell types must be consistent with reference matrix'

        self.observed_phi = observed_phi
        self.observed_sigma = observed_sigma
        self.observed_m = observed_m
        self._reset()

    def _reset(self):
        self.psis = None
        self.filter_idxs = None
        self.alpha_hats = None
        self.n_hats = None
        self.phi_hat = None
        self.marginal_ci_errs = None

    def predict(self, psis):
        assert psis.shape[0] == self.observed_phi.shape[0], 'Bulks must have the same number of genes as the reference matrix'
        self._reset()

        self.filter_idxs = np.where(psis.any(axis=1))
        alpha_hats, n_hats, phi_hat = find_mixtures(
            self.observed_phi.to_numpy()[self.filter_idxs],
            self.observed_sigma.to_numpy()[self.filter_idxs],
            self.observed_m.to_numpy(),
            psis.to_numpy()[self.filter_idxs])[0]

        ref_labels = self.observed_phi.columns.values.tolist()
        bulk_labels = psis.columns.values.tolist()
        self.psis = psis
        self.alpha_hats = pd.DataFrame(data=alpha_hats,
                                       index=bulk_labels, columns=ref_labels)
        self.n_hats = pd.DataFrame(
            n_hats, index=bulk_labels, columns=['n_hat'])
        self.phi_hat = pd.DataFrame(phi_hat, columns=ref_labels)

        return self.alpha_hats

    def compute_marginal_confidence_intervals(self, sig=.05):
        assert self.alpha_hats is not None, 'Please run predict before computing CIs'

        marginal_ci_errs = np.empty(
            (0, self.alpha_hats.shape[1]), dtype=np.float32)
        marginal_cis = []
        for i in range(self.alpha_hats.shape[0]):
            marginal_ci = compute_marginal_cis(
                self.phi_hat.to_numpy(),
                self.observed_sigma.to_numpy()[self.filter_idxs],
                self.observed_m.to_numpy(),
                self.alpha_hats.to_numpy()[i].reshape(1, -1),
                self.n_hats.to_numpy()[i],
                self.psis.to_numpy()[self.filter_idxs, i].reshape(-1, 1),
                sig
            )
            marginal_cis.append(marginal_ci)
            marginal_ci_errs = np.vstack((marginal_ci_errs, np.array(
                [(hi - low) / 2 for low, hi in marginal_ci])))

        ref_labels = self.observed_phi.columns.values.tolist()
        bulk_labels = self.psis.columns.values.tolist()
        self.marginal_ci_errs = pd.DataFrame(
            marginal_ci_errs, index=bulk_labels, columns=ref_labels)

        return marginal_cis

    def plot_proportions(self, plot_type='bar'):
        alpha_hats_melt = pd.melt(
            self.alpha_hats.reset_index(),
            id_vars=['index'],
            var_name='cell_type',
            value_name='proportion')

        if self.marginal_ci_errs is not None:
            marginal_ci_errs_melt = pd.melt(
                self.marginal_ci_errs.reset_index(),
                id_vars=['index'],
                var_name='cell_type',
                value_name='ci_err')
            alpha_hats_melt = pd.merge(
                alpha_hats_melt, marginal_ci_errs_melt, left_on=[
                    'index', 'cell_type'], right_on=[
                    'index', 'cell_type'])
            alpha_hats_melt['ci_low'] = alpha_hats_melt.apply(lambda row: np.clip(
                row['proportion'] - row['ci_err'], 0, 1), axis=1)
            alpha_hats_melt['ci_high'] = alpha_hats_melt.apply(lambda row: np.clip(
                row['proportion'] + row['ci_err'], 0, 1), axis=1)

        if plot_type == 'bar':
            bars = alt.Chart().mark_bar().encode(
                x=alt.X(
                    'cell_type:N', axis=alt.Axis(
                        title='Cell Type', labels=False)), y=alt.Y(
                    'proportion:Q', axis=alt.Axis(
                        title='Proportion')), color='cell_type:N', )

            if self.marginal_ci_errs is not None:
                error_bars = alt.Chart().mark_errorbar().encode(
                    x=alt.X(
                        'cell_type:N',
                        axis=alt.Axis(
                            title='Cell Type',
                            labels=False)),
                    y='ci_low:Q',
                    y2='ci_high:Q',
                )
                chart = alt.layer(
                    bars, error_bars, data=alpha_hats_melt).facet(
                    column=alt.Column(
                        'index:N', title='Bulk'), )
            else:
                chart = alt.layer(bars, data=alpha_hats_melt).facet(
                    column=alt.Column('index:N', title='Bulk'),
                )

        if plot_type == 'scatter':
            assert isinstance(
                alpha_hats_melt['index'].iloc[0], numbers.Number), 'Bulk labels must be quantitative'

            avg_line = alt.Chart(alpha_hats_melt).mark_line().encode(
                x=alt.X(
                    'index:Q', axis=alt.Axis(
                        title='Bulk Metric')), y=alt.Y(
                    'mean(proportion):Q', axis=alt.Axis(
                        title='Proportion')), color='cell_type:N', order='index:Q', )

            ind_scatter = alt.Chart(alpha_hats_melt).mark_point().encode(
                x=alt.X('index:Q', axis=alt.Axis(title='Bulk Metric')),
                y=alt.Y('proportion:Q', axis=alt.Axis(title='Proportion')),
                color='cell_type:N',
            )

            chart = avg_line + ind_scatter

        if plot_type == 'stacked':
            chart = alt.Chart(alpha_hats_melt).mark_bar().encode(
                x=alt.X(
                    'index:N', axis=alt.Axis(
                        title='Bulk')), y=alt.Y(
                    'sum(proportion):Q', axis=alt.Axis(
                        title='Proportion'), stack='normalize'), color='cell_type:N', )

        return chart



from rnasieve.model import RNASieveModel
from rnasieve.algo import find_mixtures
from rnasieve.preprocessing import (
    model_from_raw_counts,
    filter_droplet_to_facs,
    filter_facs_to_droplet,
    adjust_variances,
    off_simplex_filter,
    off_simplex_filter_absolute
)
from rnasieve.stats import (
    inverse_observed_fisher,
    cross_protocol_inverse_observed_fisher,
    compute_marginal_confidence_intervals
)



import numpy as np
import pandas as pd
import scipy.optimize
import scipy.stats
from collections import Counter
from functools import reduce
from rnasieve.model import RNASieveModel
from rnasieve.helper import CLIP_VALUE

# Filtering Functions


def _off_simplex_distances(means, variances, bulk):
    max_means = np.max(means, axis=1).reshape(-1, 1)
    min_means = np.min(means, axis=1).reshape(-1, 1)
    max_variances = np.max(variances, axis=1).reshape(-1, 1)
    below_min_distances = np.clip((min_means - bulk) / max_variances, 0, None)
    above_max_distances = np.clip((bulk - max_means) / max_variances, 0, None)
    return np.max(np.hstack(
        (below_min_distances, above_max_distances)), axis=1).reshape(-1, 1)


def off_simplex_filter(phi, sigma, psi, quantile):
    naive_LS = np.clip(np.linalg.lstsq(phi, psi, rcond=-1)[0], 0, None)
    n_init = np.linalg.norm(naive_LS, ord=1, axis=0)
    n_opt = scipy.optimize.minimize(lambda n: np.mean(_off_simplex_distances(
        phi, sigma, psi / n)), n_init, method='SLSQP', bounds=[(0, None)]).x
    dists = _off_simplex_distances(phi, sigma, psi / n_opt)
    filtered_idxs = np.where(dists <= np.quantile(dists, quantile))[0]
    return n_opt, filtered_idxs


def off_simplex_filter_absolute(phi, sigma, psi, threshold):
    naive_LS = np.clip(np.linalg.lstsq(phi, psi, rcond=-1)[0], 0, None)
    n_init = np.linalg.norm(naive_LS, ord=1, axis=0)
    n_opt = scipy.optimize.minimize(
        lambda n: np.mean(
            _off_simplex_distances(
                phi,
                sigma / n,
                psi / n)),
        n_init,
        method='SLSQP',
        bounds=[
            (0,
             None)]).x
    dists = _off_simplex_distances(phi, sigma / n_opt, psi / n_opt)[:, 0]
    filtered_idxs = np.where(dists <= threshold)
    return n_opt, filtered_idxs


def adjust_variances(phi, sigma, psi, threshold):
    naive_LS = np.clip(np.linalg.lstsq(phi, psi, rcond=-1)[0], 0, None)
    n_init = np.linalg.norm(naive_LS, ord=1, axis=0)
    n_opt = scipy.optimize.minimize(
        lambda n: np.mean(
            _off_simplex_distances(
                phi,
                sigma / n,
                psi / n)),
        n_init,
        method='SLSQP',
        bounds=[
            (0,
             None)]).x
    dists = _off_simplex_distances(phi, sigma / n_opt, psi / n_opt)
    dists = np.clip(dists, 1, None)
    return threshold * sigma * dists

# Empirical Protocol Filters
# Filter genes by their distance from the median in units of
# median deviations for various summary statistics.
# Individual thresholds are based on empirical trends in thirteen tissues
# of the Tabula Muris dataset.

# FACS/Droplet Filtering Helper Functions


def _fd_filter(phi, sigma, psi, md_plus=0, md_minus=np.inf):
    phi_max = np.max(phi, axis=1)
    sigma_max = np.max(sigma, axis=1)
    sigma_phi_ratios = np.max(sigma / np.clip(phi, CLIP_VALUE, None), axis=1)
    psi_phi_ratios = np.min(psi / np.clip(phi, CLIP_VALUE, None), axis=1)

    phi_max_median, phi_max_mad = np.median(
        phi_max), scipy.stats.median_absolute_deviation(phi_max, scale=1)
    sigma_max_median, sigma_max_mad = np.median(
        sigma_max), scipy.stats.median_absolute_deviation(sigma_max, scale=1)
    sp_ratios_median, sp_ratios_mad = np.median(
        sigma_phi_ratios), scipy.stats.median_absolute_deviation(sigma_phi_ratios, scale=1)
    pp_ratios_median, pp_ratios_mad = np.median(
        psi_phi_ratios), scipy.stats.median_absolute_deviation(psi_phi_ratios, scale=1)

    phi_max_non_zero_idxs = np.nonzero(phi_max)
    psi_non_zero_idxs = np.nonzero(psi[:, 0])
    phi_max_idxs = np.where(
        (phi_max_median -
         phi_max_mad *
         md_minus <= phi_max) & (
            phi_max <= phi_max_median +
            phi_max_mad *
            md_plus))
    sigma_max_idxs = np.where(
        (sigma_max_median -
         sigma_max_mad *
         md_minus <= sigma_max) & (
            sigma_max <= sigma_max_median +
            sigma_max_mad *
            md_plus))
    sp_ratios_idxs = np.where(
        (sp_ratios_median -
         sp_ratios_mad *
         md_minus <= sigma_phi_ratios) & (
            sigma_phi_ratios <= sp_ratios_median +
            sp_ratios_mad *
            md_plus))
    pp_ratios_idxs = np.where(
        psi_phi_ratios <= pp_ratios_median + pp_ratios_mad * md_plus)

    return reduce(
        np.intersect1d,
        (phi_max_non_zero_idxs,
         psi_non_zero_idxs,
         phi_max_idxs,
         sigma_max_idxs,
         sp_ratios_idxs,
         pp_ratios_idxs))


def _compute_fd_threshold(phi, sigma, psi):
    phi_max_non_zero_idxs = np.nonzero(np.max(phi, axis=1))
    psi_non_zero_idxs = np.nonzero(psi[:, 0])
    non_zero_idxs = np.intersect1d(phi_max_non_zero_idxs, psi_non_zero_idxs)
    psi_phi_ratios = np.min(
        psi[non_zero_idxs] /
        np.clip(
            phi[non_zero_idxs],
            CLIP_VALUE,
            None),
        axis=1)
    pp_ratios_median, pp_ratios_mad, pp_ratios_skew = np.median(
        psi_phi_ratios), scipy.stats.median_absolute_deviation(psi_phi_ratios, scale=1),
    scipy.stats.skew(psi_phi_ratios)

    if pp_ratios_mad / pp_ratios_median < 0.55 and pp_ratios_skew < 105:
        return 1
    return 5


def _compute_fd_max_iter(phi, sigma, psi):
    phi_max_non_zero_idxs = np.nonzero(np.max(phi, axis=1))
    psi_non_zero_idxs = np.nonzero(psi[:, 0])
    non_zero_idxs = np.intersect1d(phi_max_non_zero_idxs, psi_non_zero_idxs)
    psi_phi_ratios = np.min(
        psi[non_zero_idxs] /
        np.clip(
            phi[non_zero_idxs],
            CLIP_VALUE,
            None),
        axis=1)
    pp_ratios_median, pp_ratios_mad = np.median(
        psi_phi_ratios), scipy.stats.median_absolute_deviation(psi_phi_ratios, scale=1)

    if 0.52 < pp_ratios_mad / pp_ratios_median < 0.7 or .06 < np.quantile(
            psi_phi_ratios, 0.05) / pp_ratios_median < 0.18:
        return 1
    return 2


def filter_facs_to_droplet(phi, sigma, psi):
    """Given a FACS reference, filter genes suitably to deconvolve a droplet bulk."""
    md_plus = _compute_fd_threshold(phi, sigma, psi)
    max_iter = _compute_fd_max_iter(phi, sigma, psi)
    return _fd_filter(phi, sigma, psi, md_plus, 3), max_iter


def _df_filter(phi, sigma, psi, md_plus=0, md_minus=np.inf):
    phi_max = np.max(phi, axis=1)
    sigma_max = np.max(sigma, axis=1)
    sigma_phi_ratios = np.max(sigma / np.clip(phi, CLIP_VALUE, None), axis=1)
    psi_phi_ratios = np.min(psi / np.clip(phi, CLIP_VALUE, None), axis=1)

    phi_max_median, phi_max_mad = np.median(
        phi_max), scipy.stats.median_absolute_deviation(phi_max, scale=1)
    sigma_max_median, sigma_max_mad = np.median(
        sigma_max), scipy.stats.median_absolute_deviation(sigma_max, scale=1)
    sp_ratios_median, sp_ratios_mad = np.median(
        sigma_phi_ratios), scipy.stats.median_absolute_deviation(sigma_phi_ratios, scale=1)
    pp_ratios_median, pp_ratios_mad = np.median(
        psi_phi_ratios), scipy.stats.median_absolute_deviation(psi_phi_ratios, scale=1)

    phi_max_non_zero_idxs = np.nonzero(phi_max)
    psi_non_zero_idxs = np.nonzero(psi[:, 0])
    phi_max_idxs = np.where(
        (phi_max_median -
         phi_max_mad *
         md_minus <= phi_max) & (
            phi_max <= phi_max_median +
            phi_max_mad *
            md_plus))
    sigma_max_idxs = np.where(
        (sigma_max_median -
         sigma_max_mad *
         md_minus <= sigma_max) & (
            sigma_max <= sigma_max_median +
            sigma_max_mad *
            md_plus))
    sp_ratios_idxs = np.where(
        (sp_ratios_median -
         sp_ratios_mad *
         md_minus <= sigma_phi_ratios) & (
            sigma_phi_ratios <= sp_ratios_median +
            sp_ratios_mad *
            md_plus))
    pp_ratios_idxs = np.where(
        (pp_ratios_median -
         pp_ratios_mad *
         md_minus <= psi_phi_ratios) & (
            psi_phi_ratios <= pp_ratios_median +
            pp_ratios_mad *
            md_plus))

    return reduce(
        np.intersect1d,
        (phi_max_non_zero_idxs,
         psi_non_zero_idxs,
         phi_max_idxs,
         sigma_max_idxs,
         sp_ratios_idxs,
         pp_ratios_idxs))


def _compute_df_threshold(phi, sigma, psi):
    phi_max_non_zero_idxs = np.nonzero(np.max(phi, axis=1))
    psi_non_zero_idxs = np.nonzero(psi[:, 0])
    non_zero_idxs = np.intersect1d(phi_max_non_zero_idxs, psi_non_zero_idxs)
    sigma_phi_ratios = np.max(
        sigma[non_zero_idxs] /
        np.clip(
            phi[non_zero_idxs],
            CLIP_VALUE,
            None),
        axis=1)

    sp_ratios_skew = scipy.stats.skew(sigma_phi_ratios)

    if sp_ratios_skew <= 40:
        return 4
    elif sp_ratios_skew <= 60:
        return 1
    return 7


def _compute_df_max_iter(phi, sigma, psi):
    phi_max_non_zero_idxs = np.nonzero(np.max(phi, axis=1))
    psi_non_zero_idxs = np.nonzero(psi[:, 0])
    non_zero_idxs = np.intersect1d(phi_max_non_zero_idxs, psi_non_zero_idxs)
    psi_phi_ratios = np.min(
        psi[non_zero_idxs] /
        np.clip(
            phi[non_zero_idxs],
            CLIP_VALUE,
            None),
        axis=1)

    pp_ratios_left_tail = np.quantile(
        psi_phi_ratios, 0.05) / np.median(psi_phi_ratios)

    return 1 if pp_ratios_left_tail >= 0.1 else 3


def filter_droplet_to_facs(phi, sigma, psi):
    """Given a droplet reference, filter genes suitably to deconvolve a FACS bulk."""
    md_plus = _compute_df_threshold(phi, sigma, psi)
    max_iter = _compute_df_max_iter(phi, sigma, psi)
    return _df_filter(phi, sigma, psi, md_plus, np.inf), max_iter


def _trimmed_mean_mtx(M, frac):
    totals = M.sum(axis=0)
    sorted_idxs = np.argsort(totals)
    trim_idx = int(M.shape[1] * frac)
    return M[:, sorted_idxs[trim_idx:-trim_idx]]


def model_from_raw_counts(
        raw_counts,
        bulks,
        trim_percent=0.02,
        gene_thresh=0.2,
        normalization=True):
    """Given raw counts in the form of a dictionary { label : matrix } and a
    matrix of bulks, produce a RNASieveModel and corresponding psis.

    Matrices should be sorted by genes over the same set of genes.
    """
    g = bulks.shape[0]
    m = []
    labels = sorted(raw_counts.keys())

    phi = np.empty((g, 0), dtype=np.float32)
    sigma = np.empty((g, 0), dtype=np.float32)
    high_info_cts = Counter()

    for label in labels:
        data = _trimmed_mean_mtx(raw_counts[label], trim_percent)
        if normalization:
            data = data * 1e6 / np.sum(data, axis=0)
        high_info_cts.update(np.where(np.count_nonzero(
            data, axis=1) > gene_thresh * data.shape[1])[0])
        m.append(data.shape[1])
        phi = np.hstack((phi, np.mean(data, axis=1).reshape((-1, 1))))
        sigma = np.hstack((sigma, data.var(axis=1).reshape((-1, 1))))

    m = np.array(m).reshape(1, -1)

    # remove low info idxs
    high_info_idxs = sorted(
        [e for e in list(set(high_info_cts.elements())) if high_info_cts[e] > 0])
    phi = phi[high_info_idxs]
    sigma = sigma[high_info_idxs]
    psis = bulks[high_info_idxs]

    # remove all zero rows
    non_zero_idxs = np.where(phi.any(axis=1))
    phi = phi[non_zero_idxs]
    sigma = sigma[non_zero_idxs]
    psis = psis[non_zero_idxs]

    # convert to DataFrame
    phi_pd = pd.DataFrame(phi, columns=labels)
    sigma_pd = pd.DataFrame(sigma, columns=labels)
    m_pd = pd.DataFrame(m, columns=labels)
    psis_pd = pd.DataFrame(
        psis, columns=[
            f'Bulk {i}' for i in range(
                psis.shape[1])])

    return RNASieveModel(phi_pd, sigma_pd, m_pd), psis_pd



from distutils.core import setup
setup(
    name='rnasieve',
    packages=['rnasieve'],
    version='0.1',
    license='GPL',
    description='A library for the statistical deconvolution of RNA bulk samples with single-cell references.',
    author='Justin Hong',
    author_email='justinhong@berkeley.edu',
    url='https://github.com/songlab-cal/rna-sieve',
    download_url='https://github.com/songlab-cal/archive/v_01.tar.gz',
    keywords=[
        'rna',
        'deconvolution',
        'statistics',
        'single-cell',
        'proportion',
        'bulk'],
    install_requires=[
        'numpy',
        'cvxpy',
        'scipy',
        'pandas',
        'altair',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



"""
Helper functions for computing common terms used in RNA-Sieve.
"""

import numpy as np

# Constants

CLIP_VALUE = 1e-10
ALPHA_EPS = 1e-4
ALPHA_MAX_ITER = 1

# General Functions


def compute_mixture_sigma(alpha, sigma, phi):
    return (phi**2 + sigma) @ alpha.T - (phi @ alpha.T)**2


def compute_weighted_norm(vec, weights, p=2):
    return np.sum(vec**p * weights)


def compute_full_likelihood(
        phi, observed_phi, observed_psis, alphas, sigma, ns, m):
    mixture_variance = np.clip(compute_mixture_sigma(
        alphas, sigma, phi), CLIP_VALUE, None)
    phi_likelihood = np.sum(np.tile(m, (len(phi), 1)) /
                            sigma * (phi - observed_phi)**2)
    psi_likelihood = np.sum(ns / mixture_variance *
                            (phi @ alphas.T - observed_psis / ns)**2)
    log_term = np.sum(np.log(ns * mixture_variance))
    return phi_likelihood + psi_likelihood + log_term


def compute_row_likelihood(phi_row, observed_phi_row,
                           observed_psi_scalars, alphas, sigma_row, ns, m):
    mixture_variance = np.clip(compute_mixture_sigma(
        alphas, sigma_row, phi_row), CLIP_VALUE, None)
    phi_likelihood = np.sum(m / sigma_row * (phi_row - observed_phi_row)**2)
    psi_likelihood = np.sum(ns / mixture_variance *
                            (phi_row @ alphas.T - observed_psi_scalars / ns)**2)
    log_term = np.sum(np.log(ns * mixture_variance))
    return phi_likelihood + psi_likelihood + log_term
